@@ -4,6 +4,7 @@
 
 ;; Author: Wojciech Siewierski
 ;;         Rudolf Schlatte <rudi@constantly.at>
+;;         Kristoffer Balintona <krisbalintona@gmail.com>
 ;; URL: https://codeberg.org/emacs-jj-vc/vc-jj.el
 ;; Version: 0.4
 ;; Package-Requires: ((emacs "28.1") (compat "29.4"))
@@ -24,39 +25,45 @@
 
 ;;; Commentary:
 
-;; A backend for vc.el to handle Jujutsu (jj) repositories.
+;; This package provides a backend for vc.el to handle Jujutsu (jj)
+;; repositories.  Jujutsu uses a change-centric model that differs in
+;; important ways from Git’s commit-centric model.  Vc-jj adapts
+;; vc.el’s abstractions to match jj terminology and behavior.
 ;;
-;; Jujutsu is a distributed version control system that uses the same
-;; on-disk storage format as git (although other storage backends are
-;; possible).  Jj repositories are "co-located" by default, which
-;; means they contain both a '.jj' and a '.git' directory in the
-;; repository root.  Therefore, vc-jj can re-use some functionality of
-;; vc-git, like handling of '.gitignore' files.  It is possible to run
-;; git commands in a co-located repository, although git will complain
-;; about being in "detached HEAD" state.
+;; Users may see and customize all options by pressing
 ;;
-;; Note that 'JJ' should come before 'Git' in `vc-handled-backends' so
-;; that 'vc' chooses the jj backend.
-;;
-;; In jj terminology, vc "revisions" are called "changes" and are
-;; identified by a "change ID"; the code in vc-jj tries to adhere to
-;; this terminology.
-;;
-;; In contrast to git, there is no staging area or check-in operation:
-;; as the user modifies a file, the changes are automatically
-;; recorded.  Note that the change ID stays constant when the current
-;; change is being modified, e.g., by editing a file.  Underlying each
-;; change is a (git-type) commit with a commit ID; the commit ID
-;; changes when a change gets modified.  Most jj commands accept both
-;; change IDs and commit IDs.  We display both change ID and commit ID
-;; where appropriate, for example in the extra headers of vc-dir
-;; buffers.
+;;  'M-x customize-group vc-jj RET'
 ;;
+;; and using the Customize menu.
 
+;; STATUS AND LIMITATIONS
+;;
+;; Vc-jj implements most commonly used vc.el commands (such as
+;; `vc-dir', `vc-diff', `vc-print-log', and file state queries), but
+;; some operations are not yet supported or behave differently from
+;; vc-git due to jj’s model.
+
+;; FILE STRUCTURE
+;; 
 ;; After the "Customization" and "Internal Utilities" sections, the
 ;; organization of this file follows the "BACKEND PROPERTIES" section
-;; at the beginning of 'vc.el', implementing the functions listed
-;; there in the same order.
+;; of the preamble of the 'vc.el' file: each outline heading
+;; corresponding to a vc backend method and the contents of each
+;; heading relate to implementing that backend method.
+
+;; FEEDBACK AND CONTRIBUTIONS
+;;
+;; Bug reports and contributions are welcome, especially for
+;; unimplemented vc.el backend methods.  The home of this project is
+;; found here: https://codeberg.org/emacs-jj-vc/vc-jj.el.  Users may
+;; file bug reports in the "Issues" tab and create pull requests in
+;; the "Pull requests" tab.
+;;
+;; Vc-jj prefers jujutsu's terminology and attempts to adhere to it,
+;; so we ask contributors to try their best to use the terminology
+;; specific to jujutsu as opposed to other version control systems,
+;; such as Git.  A brief description of these differences can be found
+;; above.
 
 ;;; Code:
 
@@ -83,6 +90,88 @@
   :type 'string
   :risky t)
 
+(defcustom vc-jj-root-log-format
+  (list
+   ;; Log format (passed as the template for "jj log")
+   "
+if(root,
+  format_root_commit(self),
+  label(if(current_working_copy, \"working_copy\"),
+    concat(
+      separate(\" \",
+        change_id ++ \"​\" ++ change_id.shortest(8).prefix() ++ \"​\" ++ change_id.shortest(8).rest(),
+        if(author.name(), author.name(), if(author.email(), author.email().local(), email_placeholder)),
+        commit_timestamp(self).format(\"%Y-%m-%d\"),
+        bookmarks,
+        tags,
+        working_copies,
+        if(git_head, label(\"git_head\", \"git_head()\")),
+        format_short_commit_id(commit_id),
+        if(conflict, label(\"conflict\", \"conflict\")),
+        if(config(\"ui.show-cryptographic-signatures\").as_boolean(),
+          format_short_cryptographic_signature(signature)),
+        if(empty, label(\"empty\", \"(empty)\")),
+        if(description,
+          description.first_line(),
+          label(if(empty, \"empty\"), description_placeholder),
+        ),
+      ) ++ \"\n\",
+    ),
+  )
+)"
+   ;; Log entry regexp
+   (rx
+    line-start
+    ;; Graph
+    (+? nonl) " "
+    ;; Full change ID
+    (group (+ (any "K-Zk-z")))
+    space
+    ;; Visible change ID
+    (group (+ (any "K-Zk-z")))
+    space
+    (group (+ (any "K-Zk-z")))
+    " "
+    ;; Author
+    (group (* nonl)) " "
+    ;; Time
+    (group (= 4 (any num)) "-" (= 2 (any num)) "-" (= 2 (any num)))
+    ;; Tags and  bookmarks
+    (group (*? nonl)) " "
+    ;; Commit ID
+    (group (+ (any hex))) " "
+    ;; Special states
+    (group (opt "conflict "))
+    (group (opt "(empty) "))
+    (group (opt "(no description set)"))
+    ;; Description
+    (* nonl) line-end)
+   ;; Font lock keywords
+   '((1 '(face nil invisible t))        ; Full change ID
+     (2 'log-view-message)              ; Short change ID
+     (3 'change-log-list)               ; Rest of Change ID
+     (4 'change-log-name)               ; Author name
+     (5 'change-log-date)               ; Date
+     (6 'vc-jj-log-view-bookmark)       ; Bookmark names
+     (7 'vc-jj-log-view-commit)         ; Commit ID
+     (8 'vc-conflict-state)             ; Conflict marker
+     (9 'change-log-function)           ; No description marker
+     (10 'change-log-function)))        ; Revision description
+  "JJ log format for `vc-print-root-log'.
+This option determines the format and fontification of the JJ Log View
+buffer created from `vc-print-root-log'.
+
+It should be a list of the form (FORMAT REGEXP KEYWORDS), where FORMAT
+is a format string (a JJ template passed to \"jj log\"), REGEXP is a
+regular expression matching a single entry in the \"jj log\" output, and
+KEYWORDS is a list of font lock keywords (see
+`font-lock-keywords'and `(elisp) Search-based Fontification') for
+highlighting the Log View buffer.
+
+REGEXP may define capture groups that KEYWORDS can use to fontify
+various regions of the Log View buffer."
+  :type '(list string regexp (repeat sexp)))
+
 (defcustom vc-jj-global-switches '("--no-pager" "--color" "never")
   "Global switches to pass to any jj command."
   :type '(choice (const :tag "None" nil)
@@ -113,6 +202,14 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 		 (string :tag "Argument String")
 		 (repeat :tag "Argument List" :value ("") string)))
 
+(defface vc-jj-log-view-commit
+  '((t :weight light :inherit (shadow italic)))
+  "Face for commit IDs in `vc-jj-log-view-mode' buffers.")
+
+(defface vc-jj-log-view-bookmark
+  '((t :weight bold :inherit log-view-message))
+  "Face for bookmark names in `vc-jj-log-view-mode' buffers.")
+
 ;;; Internal Utilities
 
 ;; Note that 'JJ' should come before 'Git' in `vc-handled-backends',
@@ -133,6 +230,7 @@ When FILENAME is not inside a JJ repository, throw an error."
     (error "File is not inside a JJ repository: %s" filename)))
 
 (defun vc-jj--set-up-process-buffer (buffer root command)
+  "Prepare BUFFER for execution of COMMAND in directory ROOT."
   (with-current-buffer buffer
     (vc-run-delayed
       (vc-compilation-mode 'jj)
@@ -206,10 +304,14 @@ stderr and1 `vc-do-command' cannot separate output to stdout and stderr."
 ;;; BACKEND PROPERTIES
 
 ;;;; revision-granularity
-(defun vc-jj-revision-granularity () 'repository)
+(defun vc-jj-revision-granularity ()
+  "Jj implementation of vc property `revision-granularity'."
+  'repository)
 
 ;;;; update-on-retrieve-tag
-(defun vc-jj-update-on-retrieve-tag () nil)
+(defun vc-jj-update-on-retrieve-tag ()
+  "JJ-specific implementation of `update-on-retrieve-tag' property."
+  nil)
 
 ;;; STATE-QUERYING FUNCTIONS
 
@@ -442,7 +544,7 @@ parents.map(|c| concat(
   "Return the current change id of the repository containing FILE."
   (when-let* ((default-directory (vc-jj-root file)))
     ;; 'jj log' might print a warning at the start of its output,
-    ;; e.g., "Warning: Refused to snapshot some files".  The output we
+    ;; e.g., "Warning: Refused to snapshot some files."  The output we
     ;; want will be printed afterwards.
     (car (last (vc-jj--process-lines "log" "--no-graph"
                                      "-r" "@"
@@ -450,7 +552,9 @@ parents.map(|c| concat(
 
 ;;;; checkout-model
 
-(defun vc-jj-checkout-model (_files) 'implicit)
+(defun vc-jj-checkout-model (_files)
+  "JJ-specific implementation of `vc-checkout-model'."
+  'implicit)
 
 ;;;; mode-line-string
 
@@ -549,9 +653,9 @@ If REV is not specified, revert the file as with `vc-jj-revert'."
   "History variable for `vc-jj-pull'.")
 
 (defun vc-jj-pull (prompt)
-  "Pull changes from an upstream repository, invoked via \\[vc-update].
-Normally, this runs \"jj git fetch\".  If PROMPT is non-nil, prompt for
-the jj command to run."
+  "JJ-specific implementation of `vc-pull'.
+If PROMPT is non-nil, prompt for the jj command to run (default is \"jj
+git fetch\")."
   (let* ((command (if prompt
                       (split-string-shell-command
 		       (read-shell-command
@@ -572,9 +676,9 @@ the jj command to run."
   "History variable for `vc-jj-push'.")
 
 (defun vc-jj-push (prompt)
-  "Push changes to an upstream repository, invoked via \\[vc-push].
-Normally, this runs \"jj git push\".  If PROMPT is non-nil, prompt for
-the command to run, e.g., the semi-standard \"jj git push -c @-\"."
+  "JJ-specific implementation of `vc-push'.
+If PROMPT is non-nil, prompt for the command to run (default is \"jj git
+push\")."
   (let* ((command (if prompt
                       (split-string-shell-command
 		       (read-shell-command
@@ -594,6 +698,7 @@ the command to run, e.g., the semi-standard \"jj git push -c @-\"."
 ;;;; get-change-comment
 
 (defun vc-jj-get-change-comment (_files rev)
+  "Get the change comment of revision REV."
   (vc-jj--command-parseable "log" "--no-graph" "-n" "1"
                             "-r" rev "-T" "description"))
 
@@ -601,6 +706,7 @@ the command to run, e.g., the semi-standard \"jj git push -c @-\"."
 
 ;; TODO: protect immutable changes
 (defun vc-jj-modify-change-comment (_files rev comment)
+  "Set the change comment of revision REV to COMMENT."
   (let ((comment (car (log-edit-extract-headers () comment))))
     (vc-jj--command-dispatched nil 0 nil "desc" rev "-m" comment "--quiet")))
 
@@ -612,53 +718,43 @@ the command to run, e.g., the semi-standard \"jj git push -c @-\"."
 
 ;;;; print-log
 
-(defvar vc-jj--log-default-template
-  "
-if(root,
-  format_root_commit(self),
-  label(if(current_working_copy, \"working_copy\"),
-    concat(
-      separate(\" \",
-        change_id.shortest(8).prefix() ++ \"​\" ++ change_id.shortest(8).rest(),
-        if(author.name(), author.name(), if(author.email(), author.email().local(), email_placeholder)),
-        commit_timestamp(self).format(\"%Y-%m-%d\"),
-        bookmarks,
-        tags,
-        working_copies,
-        if(git_head, label(\"git_head\", \"git_head()\")),
-        format_short_commit_id(commit_id),
-        if(conflict, label(\"conflict\", \"conflict\")),
-        if(config(\"ui.show-cryptographic-signatures\").as_boolean(),
-          format_short_cryptographic_signature(signature)),
-        if(empty, label(\"empty\", \"(empty)\")),
-        if(description,
-          description.first_line(),
-          label(if(empty, \"empty\"), description_placeholder),
-        ),
-      ) ++ \"\n\",
-    ),
-  )
-)
-")
-
-(defun vc-jj-print-log (files buffer &optional _shortlog start-revision limit)
-  "Print commit log associated with FILES into specified BUFFER."
-  ;; FIXME: limit can be a revision string, in which case we should
-  ;; print revisions between start-revision and limit
+(defun vc-jj-print-log (files buffer &optional shortlog start-revision limit)
+  "Print commit log associated with FILES into specified BUFFER.
+If SHORTLOG is non-nil, use a short log format similar to
+`vc-jj-root-log-format'.  If START-REVISION is non-nil, it is a string
+of the newest revision in the log to show.  If LIMIT is a number, show
+no more than this many entries.  If LIMIT is a non-empty string, use it
+as a base revision."
   (vc-setup-buffer buffer)
   (let ((inhibit-read-only t)
-        (args (append
-               (and limit
-                    (list "-n" (number-to-string limit)))
-               (if start-revision
-                 (list "-r" (concat "::" start-revision))
-                 (list "-r" "::"))
-               (list "-T" vc-jj--log-default-template "--")
-               (unless (string-equal (vc-jj-root (car files)) (car files))
-                 files))))
+        (files
+         ;; There is a special case when FILES has just the root of
+         ;; the project as its only element (e.g., when calling
+         ;; `vc-print-root-log').  In this case, we do not specify a
+         ;; fileset to jj because doing do would cause the log to only
+         ;; show ancestors of START-REVISION (even if the fileset is
+         ;; "all()").  This behavior is undesirable in
+         ;; `vc-print-root-log' (in a JJ context of bookmarks), since
+         ;; users expect to see descendants as well
+         (unless (file-equal-p (vc-jj-root (car files)) (car files))
+           files))
+        (args (append (pcase limit
+                        ;; When LIMIT is a number, only show up to
+                        ;; that many revisions
+                        ((pred numberp)
+                         (list "-n" (number-to-string limit)
+                               "-r" (concat "::" start-revision)))
+                        ;; When LIMIT is a string, it is a revision.
+                        ;; In that case, show the revisions between
+                        ;; LIMIT and START-REVISION, not including the
+                        ;; revision LIMIT.
+                        ((pred stringp)
+                         (list "-r" (format "%s::%s & ~%s" limit start-revision limit))))
+                      (if shortlog
+                          (list "-T" (car vc-jj-root-log-format))
+                        (list "--no-graph" "-T" "builtin_log_detailed")))))
     (with-current-buffer buffer
-      (apply #'vc-jj--command-dispatched buffer
-        'async nil "log" args))))
+      (apply #'vc-jj--command-dispatched buffer 'async files "log" args))))
 
 ;;;; log-outgoing
 
@@ -676,43 +772,26 @@ if(root,
 
 ;;;; log-view-mode
 
-(defvar vc-jj--logline-re
-  (rx
-    line-start
-    ;; graph
-    (+? nonl)
-    " "
-    ;; change-id
-    (group (+ (any "K-Zk-z")))
-    space
-    (group (+ (any "K-Zk-z")))
-    " "
-    ;; author
-    (group (* nonl))
-    " "
-    ;; time
-    (group
-      (= 4 (any num)) "-" (= 2 (any num)) "-" (= 2 (any num)))
-      ;; " "
-      ;; (= 2 (any num)) ":" (= 2 (any num)) ":" (= 2 (any num)))
-    ;; tags, bookmarks, commit-id
-    (group
-      (*? nonl))
-
-    " "
-    ;; commit-id
-    (group
-      (+ (any hex)))
-    " "
-    ;; special states
-    (? (group (or (: "(empty) ")
-                "")))
-    (? (group (or (: "(no description set)" (* nonl))
-                "")))
-    ;; regular description
-    (* nonl)
-    line-end)
-  "Regular expression matching one line in the outpout of 'jj log'.")
+(defun vc-jj-log-view-restore-position ()
+  "Restore the position of point prior to reverting.
+When this function is added to `revert-buffer-restore-functions' in a
+`vc-jj-log-view-mode' buffer, after reverting the buffer, restore the
+position of the point to the revision the point was on prior to
+reverting.  If that revision no longer exists, do not move the point."
+  (when-let* ((rev (log-view-current-tag)))
+    (lambda ()
+      (when (re-search-forward rev nil t)
+        (goto-char (match-beginning 0))
+        (beginning-of-line)
+        ;; Report to the user that we've restored the point, otherwise
+        ;; they might think the buffer was not reverted or that
+        ;; something has gone wrong (because that would likely be the
+        ;; case with the default behavior, without this function)
+        (message "Buffer reverted and point restored to revision %s"
+                 (propertize (vc-jj--command-parseable "show" "--no-patch"
+                                                       "-r" rev
+                                                       "-T" "change_id.shortest()")
+                             'face 'log-view-message))))))
 
 (defun vc-jj--expanded-log-entry (revision)
   "Return a string of the commit details of REVISION.
@@ -729,28 +808,58 @@ Called by `log-view-toggle-entry-display' in a JJ Log View buffer."
      "--no-graph" "-T" "builtin_log_detailed")
     (buffer-string)))
 
+(defvar auto-revert-mode)
+
+;; FIXME 2025-12-07: This only reverts the parent buffer, but there
+;; more often than not multiple files in a project that could be
+;; reverted.  Do we revert those too?  (Note: there is
+;; `vc-auto-revert-mode' in Emacs 31.1.)
 (defun vc-jj--reload-log-buffers ()
-  (and vc-parent-buffer
-    (with-current-buffer vc-parent-buffer
-      (revert-buffer)))
-  (revert-buffer))
+  "Revert the log view buffer and its parent buffer.
+Revert the parent buffer of the current log view buffer then revert the
+log view buffer.
+
+The \"parent buffer\" of the log view buffer is the buffer referenced by
+the variable `vc-parent-buffer'; it is the buffer in which the log view
+buffer was created."
+  (when (derived-mode-p 'vc-jj-log-view-mode)
+    (when vc-parent-buffer
+      (with-current-buffer vc-parent-buffer
+        ;; It's only necessary to ask when `auto-revert-mode' isn't
+        ;; already enabled in the `vc-parent-buffer'.  If it is, then
+        ;; it is safe to assume the user wants that buffer reverted
+        ;; automatically
+        (revert-buffer nil auto-revert-mode)))
+    ;; Revert the log view buffer
+    (revert-buffer)))
 
 (defun vc-jj-log-view-edit-change ()
-  (interactive)
+  "Edit the jj revision at point.
+Call \"jj edit\" on the revision at point."
+  (interactive nil vc-jj-log-view-mode)
   (let ((rev (log-view-current-tag)))
     (vc-jj-retrieve-tag nil rev nil)
     (vc-jj--reload-log-buffers)))
 
 (defun vc-jj-log-view-abandon-change ()
-  (interactive)
-  ;; TODO: should probably ask for confirmation, although this would be
-  ;; different from the cli
+  "Abandon the jj revision at point.
+Call \"jj abandon\" on the revision at point."
+  (interactive nil vc-jj-log-view-mode)
   (let ((rev (log-view-current-tag)))
-    (vc-jj--command-dispatched nil 0 nil "abandon" rev "--quiet")
-    (vc-jj--reload-log-buffers)))
+    (when (y-or-n-p (format "Abandon revision %s?"
+                            (propertize
+                             (vc-jj--command-parseable "show" "--no-patch"
+                                                       "-r" rev
+                                                       "-T" "change_id.shortest()")
+                             'face 'log-view-message)))
+      (vc-jj--command-dispatched nil 0 nil "abandon" rev "--quiet")
+      (vc-jj--reload-log-buffers))))
 
 (defun vc-jj-log-view-new-change ()
-  (interactive)
+  "Create a new, empty change on the revision at point.
+Call \"jj new\", creating a new revision whose parent is the revision at
+point."
+  (interactive nil vc-jj-log-view-mode)
   (let ((rev (log-view-current-tag)))
     (vc-jj--command-dispatched nil 0 nil "new" rev "--quiet")
     (vc-jj--reload-log-buffers)))
@@ -772,14 +881,14 @@ user first."
            ;; currently on, this means that the bookmark will be moved
            ;; sideways or backwards
            (backwards-move-p
-            (when (not new-bookmark-p)
-              (let* ((bookmark-rev
-                      (car (vc-jj--process-lines "show" bookmark "--no-patch"
-                                                 "-T" "self.change_id().shortest() ++ \"\n\"")))
-                     (bookmark-descendants
-                      (vc-jj--process-lines "log" "--no-graph" "-r" (concat bookmark-rev "..")
-                                            "-T" "self.change_id().shortest() ++ \"\n\"")))
-                (not (member target-rev bookmark-descendants))))))
+            (unless new-bookmark-p
+              ;; If `vc-jj--process-lines' returns nil, then that
+              ;; means the jj command returned no revisions.  This
+              ;; REVSET is the intersection between BOOKMARK, its
+              ;; descendants, and TARGET-REV.  That intersection is
+              ;; non-empty only when TARGET-REV is BOOKMARK or a
+              ;; descendant or BOOKMARK
+              (not (vc-jj--process-lines "log" "-r" (format "%s & %s::" target-rev bookmark))))))
       (when backwards-move-p
         (unless (yes-or-no-p
                  (format-prompt "Moving bookmark %s to revision %s would move it either backwards or sideways. Is this okay?"
@@ -799,8 +908,17 @@ rename."
   (when (derived-mode-p 'vc-jj-log-view-mode)
     (let* ((target-rev (log-view-current-tag))
            (bookmarks-at-rev
-            (vc-jj--process-lines "bookmark" "list" "-r" target-rev
-                                  "-T" "if(!self.remote(), self.name() ++ \"\n\")"))
+            (or (vc-jj--process-lines "bookmark" "list" "-r" target-rev
+                                      "-T" "if(!self.remote(), self.name() ++ \"\n\")")
+                ;; FIXME(KrisB 2025-12-09): Is there a more
+                ;; idiomatic/cleaner way to exit with a message than a
+                ;; `user-error' in the middle of a let binding?
+                (user-error "No bookmarks at %s"
+                            (propertize
+                             (vc-jj--command-parseable "show" "--no-patch"
+                                                       "-r" target-rev
+                                                       "-T" "change_id.shortest()")
+                             'face 'log-view-message))))
            (bookmark-old
             (if (= 1 (length bookmarks-at-rev))
                 (car bookmarks-at-rev)
@@ -833,33 +951,67 @@ delete."
       (apply #'vc-jj--command-dispatched nil 0 nil "--quiet" "bookmark" "delete" bookmarks)
       (revert-buffer))))
 
+(defvar vc-jj-log-view-mode-map
+  (let ((map (make-sparse-keymap)))
+    (keymap-set map "r" #'vc-jj-log-view-edit-change)
+    (keymap-set map "x" #'vc-jj-log-view-abandon-change)
+    (keymap-set map "i" #'vc-jj-log-view-new-change)
+    (keymap-set map "b s" #'vc-jj-log-view-bookmark-set)
+    (keymap-set map "b r" #'vc-jj-log-view-bookmark-rename)
+    (keymap-set map "b D" #'vc-jj-log-view-bookmark-delete)
+    map)
+  "Keymap for `vc-jj-log-view-mode'.")
+
 (define-derived-mode vc-jj-log-view-mode log-view-mode "JJ-Log-View"
+  "Log View mode specific for JJ."
+  :keymap vc-jj-log-view-mode-map
+
   (require 'add-log) ;; We need the faces add-log.
   ;; Don't have file markers, so use impossible regexp.
   (setq-local log-view-file-re regexp-unmatchable)
   (setq-local log-view-per-file-logs nil)
-  (setq-local log-view-message-re vc-jj--logline-re)
+  ;; The `log-view-message-re' is a regexp matching a revision.  Its
+  ;; first match group must match the revision number itself
+  (setq-local log-view-message-re
+              (if (not (memq vc-log-view-type '(long log-search with-diff)))
+                  (cadr vc-jj-root-log-format)
+                (rx bol "Commit ID: " (1+ (any alnum)) "\n"
+                    "Change ID: " (group (1+ (any alpha))))))
   ;; Allow expanding short log entries.
-  (setq truncate-lines t)
-  (setq-local log-view-expanded-log-entry-function
-    'vc-jj--expanded-log-entry)
+  (when (memq vc-log-view-type '(short log-outgoing log-incoming mergebase))
+    (setq truncate-lines t)
+    (setq-local log-view-expanded-log-entry-function 'vc-jj--expanded-log-entry))
+  ;; Fontify according to regexp capture groups (for "short" format
+  ;; log views, the relevant regexp is found in
+  ;; `vc-jj-root-log-format')
   (setq-local log-view-font-lock-keywords
-    `((,vc-jj--logline-re
-        (1 'log-view-message)
-        (2 'change-log-list)
-        (3 'change-log-name)
-        (4 'change-log-date)
-        (5 'change-log-file)
-        (6 'change-log-list)
-        (7 'change-log-function)
-        (8 'change-log-function))))
-
-  (keymap-set vc-jj-log-view-mode-map "r" #'vc-jj-log-view-edit-change)
-  (keymap-set vc-jj-log-view-mode-map "x" #'vc-jj-log-view-abandon-change)
-  (keymap-set vc-jj-log-view-mode-map "i" #'vc-jj-log-view-new-change)
-  (keymap-set vc-jj-log-view-mode-map "b s" #'vc-jj-log-view-bookmark-set)
-  (keymap-set vc-jj-log-view-mode-map "b r" #'vc-jj-log-view-bookmark-rename)
-  (keymap-set vc-jj-log-view-mode-map "b D" #'vc-jj-log-view-bookmark-delete))
+              (if (not (memq vc-log-view-type '(long log-search with-diff)))
+                  (list (cons (nth 1 vc-jj-root-log-format)
+                              (nth 2 vc-jj-root-log-format)))
+                `((,log-view-message-re
+                   (1 'change-log-acknowledgment))
+                  (,(rx bol "Commit ID" (0+ (any space)) ": " (group (1+ (any alnum))) "\n")
+                   (1 'vc-jj-log-view-commit))
+                  (,(rx bol "Bookmarks" (0+ (any space)) ": " (group (1+ not-newline)) "\n")
+                   (1 'vc-jj-log-view-bookmark))
+                  (,(rx bol (or "Author" "Committer") (0+ (any space)) ": "
+                        ;; Name
+                        (group (+? anything))
+                        ;; Email (based on `goto-address-mail-regexp')
+                        (group " <"
+                               (+ (or alnum (any "-=._+")))
+                               "@"
+                               (+ (seq (+ (or alnum (any "-_"))) "."))
+                               (+ alnum)
+                               "> ")
+                        ;; Date and time
+                        "(" (group (+ (or digit space (any "-:")))) ")\n")
+                   (1 'change-log-name)
+                   (2 'change-log-email)
+                   (3 'change-log-date)))))
+  
+  (when (boundp 'revert-buffer-restore-functions) ; Emacs 30.1
+    (add-hook 'revert-buffer-restore-functions #'vc-jj-log-view-restore-position nil t)))
 
 ;;;; show-log-entry
 
@@ -881,21 +1033,39 @@ delete."
 ;;;; diff
 
 (defun vc-jj-diff (files &optional rev1 rev2 buffer _async)
-  "Display diffs for FILES between revisions REV1 and REV2."
+  "Display diffs for FILES between revisions REV1 and REV2.
+FILES is a list of file paths.  REV1 and REV2 are the full change IDs of
+two revisions.  REV1 is the earlier revision and REV2 is the later
+revision.
+
+When BUFFER is non-nil, it is the buffer object or name to insert the
+diff into.  Otherwise, when nil, insert the diff into the *vc-diff*
+buffer.  If _ASYNC is non-nil, run asynchronously.  This is currently
+unsupported."
   ;; TODO: handle async
   (setq buffer (get-buffer-create (or buffer "*vc-diff*"))
         files (mapcar #'vc-jj--filename-to-fileset files))
   (cond
    ((not (or rev1 rev2))
-    (setq rev1 "@-"))
+    ;; Use `vc-jj-previous-revision' instead of "@-" because the
+    ;; former handles edge cases like e.g. multiple parents
+    (setq rev1 (vc-jj-previous-revision nil "@")))
    ((null rev1)
     (setq rev1 "root()")))
   (setq rev2 (or rev2 "@"))
   (let ((inhibit-read-only t)
-        (args (append (vc-switches 'jj 'diff) (list "--") files)))
+        ;; When REV1 and REV2 are the same revision, "-f REV1 -t REV2"
+        ;; (erroneously) returns an empty diff.  So we check for that
+        ;; case and use "-r REV1" instead, which returns the correct
+        ;; diff
+        (args (append (if (string= rev1 rev2)
+                          (list "-r" rev1)
+                        (list "-f" rev1 "-t" rev2))
+                      (vc-switches 'jj 'diff)
+                      (list "--") files)))
     (with-current-buffer buffer
       (erase-buffer))
-    (apply #'call-process vc-jj-program nil buffer nil "diff" "--from" rev1 "--to" rev2 args)
+    (apply #'call-process vc-jj-program nil buffer nil "diff" args)
     (if (seq-some (lambda (line) (string-prefix-p "M " line))
                   (apply #'vc-jj--process-lines "diff" "--summary" "--" files))
         1
@@ -986,6 +1156,9 @@ four groups: change id, author, datetime, line number.")
 ;;;; region-history
 
 (defun vc-jj-region-history (file buffer lfrom lto)
+  "JJ-specific version of `vc-region-history'.
+Insert into BUFFER the history (log comments and diffs) of the content
+of FILE between lines LFROM and LTO."
   ;; Support for vc-region-history via the git version which works
   ;; fine at least when co-located with git.
   (vc-git-region-history file buffer lfrom lto))
@@ -1058,10 +1231,12 @@ For jj, modify `.gitignore' and call `jj untrack' or `jj track'."
 ;;;; previous-revision
 
 (defun vc-jj-previous-revision (file rev)
-  "JJ-specific version of `vc-previous-revision'."
+  "JJ-specific version of `vc-previous-revision'.
+Return the revision number that precedes REV for FILE, or nil if no such
+revision exists."
   (if file
       (vc-jj--command-parseable "log" "--no-graph" "--limit" "1"
-                                "-r" (concat "ancestors(" rev ")")
+                                "-r" (format "ancestors(%s) & ~%s" rev rev)
                                 "-T" "change_id"
                                 "--" (vc-jj--filename-to-fileset file))
     ;; The jj manual states that "for merges, [first_parent] only
@@ -1077,7 +1252,9 @@ For jj, modify `.gitignore' and call `jj untrack' or `jj track'."
 ;;;; next-revision
 
 (defun vc-jj-next-revision (file rev)
-  "JJ-specific version of `vc-next-revision'."
+  "JJ-specific version of `vc-next-revision'.
+Return the revision that follows REV for FILE, or nil if no such
+revision exists."
   (if file
       (vc-jj--command-parseable "log" "--no-graph" "--limit" "1"
                                 "-r" (concat "descendants(" rev ")")
